from typing import Optional


def _login_request_info(username: str, password: str, scope: Optional[str] = None) -> dict:
    return {
        "url": f"https://auth.algoralabs.com/auth/realms/production/protocol/openid-connect/token",
        "data": {
            "client_id": "algora-api",
            "scope": scope,
            "grant_type": "password",
            "username": username,
            "password": password
        }
    }


def _refresh_token_request_info(refresh_token: Optional[str] = None) -> dict:
    return {
        "url": f"https://auth.algoralabs.com/auth/realms/production/protocol/openid-connect/token",
        "data": {
            "client_id": "algora-api",
            "grant_type": "refresh_token",
            "refresh_token": refresh_token
        }
    }



from typing import List, Optional, Union

from algora.common.base import Base


class LogicalDisplayName(Base):
    display_name: str
    logical_name: str


class DateRange(Base):
    start_date: str
    end_date: str
    enabled: bool


class APIFieldFilter(Base):
    logical_display: LogicalDisplayName
    operator: str  # operator can be "NOT_IN" or "IN" or "NOT_EQUAL" or "EQUAL" or "GTE" or "GT" or "LTE" or "LT"
    selected_values: List[str]


class FieldFilter(Base):
    field: str
    operator: str  # operator can be "NOT_IN" or "IN" or "NOT_EQUAL" or "EQUAL" or "GTE" or "GT" or "LTE" or "LT"
    selected_values: List[str]


class DataFilter(Base):
    date_range: Optional[DateRange]
    filters: List[Union[FieldFilter, APIFieldFilter]]



from algora.api.service.datasets.field_group.asynchronous import (
    async_get_field_group, async_get_field_groups, async_create_field_group, async_update_field_group,
    async_delete_field_group
)
from algora.api.service.datasets.field_group.synchronous import (
    get_field_group, get_field_groups, create_field_group, update_field_group, delete_field_group
)



"""
Data API query functions for timeseries data.
"""
from algora.api.data.query.asynchronous import async_query_dataset, async_query_dataset_csv, async_query_timeseries
from algora.api.data.query.synchronous import query_dataset, query_dataset_csv, query_timeseries



"""
Backtest state API.
"""
from algora.api.service.backtest.state.asynchronous import (
    async_get_all_portfolio_state, async_get_all_cash_payments, async_create_portfolio_state, async_create_cash_payment
)
from algora.api.service.backtest.state.synchronous import (
    get_all_portfolio_state, get_all_cash_payments, create_portfolio_state, create_cash_payment
)



"""
Common types.
"""
from datetime import date, datetime
from typing import NewType, Union, List, Dict

from pandas import DataFrame

Datetime = NewType('Datetime', Union[date, datetime])
DataResponse = NewType('DataResponse', Union[DataFrame, List[DataFrame], Dict[str, DataFrame], float])
PriceResult = NewType('PriceResult', Union[float])
RiskResult = NewType('RiskResult', Union[float])



"""
Data API query functions for [IEX](https://iexcloud.io/) energy data.
"""
from algora.api.data.iex.energy.asynchronous import async_historical_oil_prices
from algora.api.data.iex.energy.synchronous import historical_oil_prices



import io
from typing import Dict, Any, Optional

import pandas as pd
from pandas import DataFrame

from algora.api.service.data_engine.__util import _analyze_data_request_info, _transform_data_request_info
from algora.api.service.data_engine.model import TransformOverride
from algora.common.decorators.data import async_data_request
from algora.common.function import no_transform
from algora.common.requests import __async_post_request


@async_data_request(transformer=no_transform)
async def async_analyze_data(data: pd.DataFrame) -> Dict[str, Any]:
    request_info = _analyze_data_request_info(data)
    return await __async_post_request(**request_info)


@async_data_request(processor=lambda r: r.content, transformer=lambda r: pd.read_parquet(io.BytesIO(r)))
async def async_transform_data(data: pd.DataFrame, transform_override: Optional[TransformOverride] = None) -> DataFrame:
    request_info = _transform_data_request_info(data, transform_override)
    return await __async_post_request(**request_info)



from typing import Dict, Any

from algora.api.service.research.resource.__util import _get_resource_request_info, _delete_resource_request_info
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import __async_get_request, __async_delete_request


@async_data_request(transformer=lambda data: data, processor=lambda response: response.content)
async def async_get_resource(id: str) -> Dict[str, Any]:
    """
    Asynchronously get resource by ID.

    Args:
        id (str): Resource ID

    Returns:
        Dict[str, Any]: Resource response
    """
    request_info = _get_resource_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_resource(id: str) -> None:
    """
    Asynchronously delete resource by ID.

    Args:
        id (str): Resource ID

    Returns:
        None
    """
    request_info = _delete_resource_request_info(id)
    return await __async_delete_request(**request_info)



"""
Data classes for API requests.
"""
from datetime import datetime
from typing import List, Optional

from pydantic import Field

from algora.common.base import Base
from algora.common.enum import Order
from algora.common.type import Datetime


class TimeseriesQueryRequest(Base):
    id: Optional[str] = Field(default=None)
    dataset_id: Optional[str] = Field(default=None)
    asset_id: Optional[str] = Field(default=None)
    date: Optional[Datetime] = Field(default=None)
    start_date: Optional[datetime] = Field(default=None)
    end_date: Optional[datetime] = Field(default=None)
    as_of: Optional[datetime] = Field(default=None)
    page: Optional[int] = Field(default=None)
    limit: Optional[int] = Field(default=None)
    fields: Optional[List[str]] = Field(default=None)
    sort: Order = Field(default=Order.DESC)
    where: Optional[List[dict]] = Field(default=None)



from algora.common.base import BaseEnum


class AssetClass(BaseEnum):
    COMMODITY = "commodity"
    CREDIT = "credit"
    EQUITY = "equity"
    FOREX = "forex"
    RATES = "rates"


class Repository(BaseEnum):
    CME = "CME"
    DTCC = "DTCC"
    ICE = "ICE"



from datetime import date
from typing import Optional

from algora.api.service.backtest.state.enum import BacktestEventType
from algora.common.base import Base


class BacktestEvent(Base):
    backtest_id: str
    backtest_type: BacktestEventType


class CashPaymentRequest(Base):
    backtest_id: str
    payment_date: date
    type: str
    amount: float
    position_id: Optional[str]
    currency: Optional[str]


class PortfolioStateRequest(Base):
    backtest_id: str
    date: date
    valuation: float
    pnl: float
    portfolio: str


class CashPaymentResponse(BacktestEvent):
    id: str
    backtest_id: str
    backtest_type: BacktestEventType
    payment_date: date
    type: str
    amount: float
    position_id: Optional[str]
    currency: Optional[str]
    created_at: int


class PortfolioStateResponse(BacktestEvent):
    id: str
    backtest_id: str
    backtest_type: BacktestEventType
    date: date
    valuation: float
    pnl: float
    portfolio: str
    created_at: int


class BacktestMetricRequest(Base):
    backtest_id: str
    metrics: str



"""
Data API query functions for SDR data.
"""
from algora.api.data.sdr.asynchronous import (
    async_get_by_date, async_get_distinct_in_field, async_commodity, async_credit, async_equity, async_forex,
    async_rates
)
from algora.api.data.sdr.synchronous import (
    get_by_date, get_distinct_in_field, commodity, credit, equity, forex, rates
)



from typing import Optional, Dict, Union, Any

from algora.common.base import Base


class DocumentRequest(Base):
    name: str
    type: str
    classification: str
    document: Union[str, Dict]


class SearchDocumentRequest(Base):
    name: Optional[str]
    type: Optional[str]
    classification: Optional[str]
    is_released: Optional[bool]


class Document(Base):
    pk: str
    parent_pk: str
    original_pk: str
    name: str
    type: str
    classification: str
    document: Union[str, Dict[str, Any]]
    created_by: str
    created_at: int
    updated_by: str
    updated_at: int
    released_by: str
    released_at: int
    is_released: bool
    deleted_by: str
    deleted_at: int
    is_deleted: bool



import pandas as pd
from requests import Response

from algora.api.data.query.__util import _query_datasets_request_info, _query_dataset_csv_request_info, \
    _query_timeseries_request_info
from algora.api.data.query.model import TimeseriesQueryRequest
from algora.common.decorators import data_request
from algora.common.requests import __async_post_request, __async_get_request


async def async_query_dataset(id: str, data=None, json=None) -> Response:
    """
    Asynchronously query dataset by ID.

    Args:
        id (str): Dataset ID
        data (Any): Data to POST
        json (Any): Data to POST

    Returns:
        Response: HTTP response object
    """
    request_info = _query_datasets_request_info(id, data, json)
    return await __async_post_request(**request_info)


async def async_query_dataset_csv(id: str, data=None) -> Response:
    """
    Asynchronously query dataset CSV by ID.

    Args:
        id (str): Dataset ID
        data (Any): Data to POST

    Returns:
        Response: HTTP response object
    """
    request_info = _query_dataset_csv_request_info(id, data)
    return await __async_get_request(**request_info)


@data_request
def async_query_timeseries(request: TimeseriesQueryRequest) -> pd.DataFrame:
    """
    Asynchronously query timeseries dataset by timeseries query request.

    Args:
        request (TimeseriesQueryRequest): Timeseries query request

    Returns:
        pd.DataFrame: Timeseries DataFrame
    """
    request_info = _query_timeseries_request_info(request)
    return __async_post_request(**request_info)



from algora.api.service.backtest.state.model import PortfolioStateRequest, CashPaymentRequest
from algora.common.enum import Order


def _get_all_portfolio_state_request_info(backtest_id: str, order: Order = Order.ASC, **kwargs) -> dict:
    config = {
        "endpoint": f"config/backtest/{backtest_id}/state/portfolio?order={order}"
    }
    kwargs.update(config)

    return kwargs


def _get_all_cash_payments_request_info(backtest_id: str, order: Order, **kwargs) -> dict:
    config = {
        "endpoint": f"config/backtest/{backtest_id}/state/cash_payment?order={order}"
    }
    kwargs.update(config)

    return kwargs


def _create_portfolio_state_request_info(request: PortfolioStateRequest) -> dict:
    return {
        "endpoint": "config/backtest/state/portfolio",
        "json": request.request_dict()
    }


def _create_cash_payment_request_info(request: CashPaymentRequest) -> dict:
    return {
        "endpoint": "config/backtest/state/cash_payment",
        "json": request.request_dict()
    }



from pandas import DataFrame

from algora.api.data.iex.__util import __async_base_request
from algora.api.data.iex.energy.__util import __build_params
from algora.common.decorators.data import async_data_request


@async_data_request
async def async_historical_oil_prices(symbol: str, **kwargs) -> DataFrame:
    """
    Asynchronous wrapper for IEX historical energy prices via
    [Time Series API](https://iexcloud.io/docs/api/#time-series-endpoint).

    Args:
        symbol (str): "DCOILWTICO" or "DCOILBRENTEU"
        **kwargs: Optional args to pass to the IEX API

    Returns:
        DataFrame: Historical oil prices for the provided symbol
    """
    # default query params
    params = __build_params(**kwargs)
    return await __async_base_request(f"time-series/energy/{symbol}", **params)



"""
Backtest API.
"""
from algora.api.service.backtest.asynchronous import (
    async_get_backtest, async_get_backtests, async_create_backtest, async_update_backtest, async_delete_backtest
)
from algora.api.service.backtest.synchronous import (
    get_backtest, get_backtests, create_backtest, update_backtest, delete_backtest
)



from datetime import date
from typing import Optional

from pydantic import Field, validator

from algora.common.base import Base


class FredQuery(Base):
    api_key: str
    series_id: str
    file_type: str = Field(default='json')
    observation_start: Optional[date] = None
    observation_end: Optional[date] = None

    @validator("observation_start")
    def observation_start_to_string(cls, d):
        # necessary for serialization
        return d.isoformat()

    @validator("observation_end")
    def observation_end_to_string(cls, d):
        # necessary for serialization
        return d.isoformat()



from typing import Optional

import aiohttp
from requests import Response

from algora.common.config import EnvironmentConfig
from algora.common.decorators import async_authenticated_request
from algora.common.function import build_response_obj

config = EnvironmentConfig()


@async_authenticated_request
async def __async_get_request(
        endpoint: str,
        url_key: str = "algora",
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        timeout: int = 30
) -> Response:
    """
    Wrapper method for asynchronous get requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(
                url=f"{config.get_url(url_key)}/{endpoint}",
                headers=headers or {},
                params=params,
                timeout=timeout
        ) as response:
            result = await build_response_obj(response)

    return result


@async_authenticated_request
async def __async_put_request(
        endpoint: str,
        url_key: str = "algora",
        data=None,
        json=None,
        files=None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        timeout: int = 30
) -> Response:
    """
    Wrapper method for asynchronous put requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        data (Any): Data passed to the API
        json (Any): JSON passed to the API
        files (Any): Files passed to the API
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    data = data if data is None else data.update({"files": files})
    async with aiohttp.ClientSession() as session:
        async with session.put(
                url=f"{config.get_url(url_key)}/{endpoint}",
                data=data,
                json=json,
                headers=headers or {},
                params=params,
                timeout=timeout
        ) as response:
            result = await build_response_obj(response)

    return result


@async_authenticated_request
async def __async_post_request(
        endpoint: str,
        url_key: str = "algora",
        files=None,
        data=None,
        json=None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        timeout: int = 30
) -> Response:
    """
    Wrapper method for asynchronous post requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        data (Any): Data passed to the API
        json (Any): JSON passed to the API
        files (Any): Files passed to the API
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    data = data if data is None else data.update({"files": files})
    async with aiohttp.ClientSession() as session:
        async with session.post(
                url=f"{config.get_url(url_key)}/{endpoint}",
                data=data,
                json=json,
                headers=headers or {},
                params=params,
                timeout=timeout
        ) as response:
            result = await build_response_obj(response)

    return result


@async_authenticated_request
async def __async_delete_request(
        endpoint: str,
        url_key: str = "algora",
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        timeout: int = 30
) -> Response:
    """
    Wrapper method for asynchronous delete requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    async with aiohttp.ClientSession() as session:
        async with session.delete(
                url=f"{config.get_url(url_key)}/{endpoint}",
                headers=headers or {},
                params=params,
                timeout=timeout
        ) as response:
            result = await build_response_obj(response)

    return result



from algora.api.service.backtest.enum import BacktestStatus
from algora.api.service.backtest.model import BacktestRequest


def _get_backtest_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/backtest/{id}"
    }


def _get_backtests_request_info() -> dict:
    return {
        "endpoint": f"config/backtest"
    }


def _create_backtest_request_info(request: BacktestRequest) -> dict:
    return {
        "endpoint": "config/backtest",
        "json": request.request_dict()
    }


def _update_backtest_request_info(id: str, status: BacktestStatus):
    return {
        "endpoint": f"config/backtest/{id}",
        "json": status.value
    }


def _delete_backtest_request_info(id: str):
    return {
        "endpoint": f"config/backtest/{id}",
    }



from algora.api.service.research.runner.__util import _get_or_create_runner_request_info
from algora.common.decorators import async_data_request
from algora.common.requests import __async_put_request


@async_data_request(transformer=lambda data: data)
async def async_get_or_create_runner() -> str:
    """
    Asynchronously get or create runner.

    Returns:
        str: Runner ID
    """
    request_info = _get_or_create_runner_request_info()
    return await __async_put_request(**request_info)



from typing import Optional, List

from algora.common.base import Base
from algora.common.enum import FieldType


class FieldRequest(Base):
    display_name: str
    logical_name: str
    type: FieldType
    width: int
    editable: bool
    hidden: bool
    display_order: int
    tags: List[str]  # TODO make enum
    schema_id: str
    field_group_id: Optional[str]



from algora.api.service.datasets.field.model import FieldRequest


def _get_field_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/field/{id}"
    }


def _create_field_request_info(request: FieldRequest) -> dict:
    return {
        "endpoint": "config/datasets/field",
        "json": request.request_dict()
    }


def _update_field_request_info(id: str, request: FieldRequest):
    return {
        "endpoint": f"config/datasets/field/{id}",
        "json": request.request_dict()
    }


def _delete_field_request_info(id: str):
    return {
        "endpoint": f"config/datasets/field/{id}",
    }



"""
Authenticated request decorators.
"""
from algora.common.decorators.authorization.asynchronous import async_authenticated_request
from algora.common.decorators.authorization.synchronous import authenticated_request



"""
Permission API requests.
"""
from typing import Dict, Any, List

from algora.api.service.permission.__util import (
    _get_permission_request_info, _delete_permission_request_info, _update_permission_request_info,
    _create_permission_request_info, _get_permissions_by_resource_id_request_info,
    _get_permission_by_resource_id_request_info
)
from algora.api.service.permission.model import PermissionRequest
from algora.common.decorators.data import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_get_request, __async_delete_request, __async_post_request, __async_put_request
)


@async_data_request(transformer=no_transform)
async def async_get_permission(id: str) -> Dict[str, Any]:
    """
    Asynchronously get permission by ID.

    Args:
        id (str): Permission ID

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _get_permission_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_permission_by_resource_id(resource_id: str) -> Dict[str, Any]:
    """
    Asynchronously get permission by resource ID.

    Args:
        resource_id (str): Resource ID

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _get_permission_by_resource_id_request_info(resource_id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_permissions_by_resource_id(resource_id: str) -> List[Dict[str, Any]]:
    """
    Asynchronously get all permissions by resource ID.

    Args:
        resource_id (str): Resource ID

    Returns:
        List[Dict[str, Any]]: List of permission response
    """
    request_info = _get_permissions_by_resource_id_request_info(resource_id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_permission(request: PermissionRequest) -> Dict[str, Any]:
    """
    Asynchronously create permission.

    Args:
        request (PermissionRequest): Permission request

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _create_permission_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_permission(id: str, request: PermissionRequest) -> Dict[str, Any]:
    """
    Asynchronously update permission.

    Args:
        id: (str): Permission ID
        request (PermissionRequest): Permission request

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _update_permission_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_permission(id: str) -> None:
    """
    Asynchronously delete permission by ID.

    Args:
        id: (str): Permission ID

    Returns:
        None
    """
    request_info = _delete_permission_request_info(id)
    return await __async_delete_request(**request_info)



"""
Permission API.
"""
from algora.api.service.permission.asynchronous import (
    async_get_permission, async_get_permission_by_resource_id, async_get_permissions_by_resource_id,
    async_create_permission, async_update_permission, async_delete_permission
)
from algora.api.service.permission.synchronous import (
    get_permission, get_permission_by_resource_id, get_permissions_by_resource_id, create_permission, update_permission,
    delete_permission
)



from algora.api.data.query.model import TimeseriesQueryRequest


def _query_datasets_request_info(id: str, data=None, json=None) -> dict:
    return {
        "endpoint": f"data/datasets/query/{id}",
        "data": data,
        "json": json
    }


def _query_dataset_csv_request_info(id: str, data=None) -> dict:
    return {
        "endpoint": f"data/datasets/query/{id}.csv",
        "data": data
    }


def _query_timeseries_request_info(request: TimeseriesQueryRequest) -> dict:
    return {
        "endpoint": f"data/timeseries",
        "json": request.request_dict()
    }



def _symbols_request_info() -> dict:
    return {
        "extension": f"ref-data/symbols"
    }


def _historical_prices_request_info(*symbol: str, **kwargs) -> dict:
    request_info = {
        'range': '1m',
        'sort': 'asc'
    }
    request_info.update(kwargs)

    if len(symbol) > 1:
        request_info.update({
            'extension': "stock/market/batch",
            'symbols': ','.join(symbol),
            'types': 'chart'
        })
    else:
        request_info.update({'extension': f"time-series/HISTORICAL_PRICES/{symbol[0]}"})

    return request_info


def _news_request_info(symbol: str, kwargs):
    request_info = {
        "extension": f"time-series/news/{symbol}"
    }
    request_info.update(kwargs)
    return request_info


def _peer_group_request_info(symbol: str) -> dict:
    return {
        "extension": f"stock/{symbol}/peers"
    }



from typing import Optional, Union, List

from algora.api.service.data_engine.enum import FieldFill
from algora.common.base import Base
from algora.common.enum import FieldType


class FieldMetric(Base):
    name: str
    type: FieldType
    length: int
    num_null: int
    num_zero: int
    min: Optional[Union[float, int]]
    max: Optional[Union[float, int]]
    std_dev: Optional[float]


class FieldOverride(Base):
    name: str
    type: Optional[FieldType] = None
    fill: Optional[FieldFill] = None


class TransformOverride(Base):
    fields: List[FieldOverride] = []
    default_fill: FieldFill = FieldFill.PREVIOUS



from typing import Callable, Union

from pydantic import Field
from typing_extensions import Literal, Annotated

from algora.common.base import Base


def serializable(
        _cls: Base = None
) -> Union[Callable, Base]:
    """
    A decorator that sets the descriptor field for a serializable class.

    NOTE: This is used for class contained in Annotations (look at the create_annotation method)

    Args:
        _cls (object): The class being decorated

    Returns:
        cls: The updated class with the descriptor field
    """

    def wrap(cls):
        annotations = {"descriptor": Literal[cls.cls_name()]}
        return type(cls.cls_name(), (cls,), {'__annotations__': annotations, 'descriptor': Field(cls.cls_name())})

    if _cls is None:
        return wrap

    return wrap(_cls)


def create_annotation(*types):
    """
    Creates an annotation used in type hinting that contains classes that are decorated with `@serializable`.

    Args:
        *types: A tuple of all the serializable class types

    Returns:
        A annotation class used in type hinting
    """

    class Annotation:
        types = ()

        @classmethod
        def annotation(cls):
            return Annotated[Union[cls.types], Field(discriminator='descriptor')]

        @classmethod
        def add_types(cls, *new_types):
            updated_types = cls.types + new_types
            cls.types = updated_types

    annotation = Annotation()
    annotation.add_types(*types)
    return Annotation



from typing import List, Dict, Any

from algora.api.service.datasets.field.model import FieldRequest
from algora.api.service.datasets.schema.__util import (
    _delete_schema_request_info, _update_schema_fields_request_info, _update_schema_request_info,
    _create_schema_request_info, _get_schema_fields_request_info, _get_schemas_request_info, _get_schema_request_info
)
from algora.api.service.datasets.schema.model import SchemaRequest
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_get_request, __async_post_request, __async_put_request, __async_delete_request
)


@async_data_request(transformer=no_transform)
async def async_get_schema(id: str) -> Dict[str, Any]:
    """
    Asynchronously get schema by ID.

    Args:
        id (str): Schema ID

    Returns:
        Dict[str, Any]: Schema response
    """
    request_info = _get_schema_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_schemas() -> List[Dict[str, Any]]:
    """
    Asynchronously get all schemas.

    Returns:
        List[Dict[str, Any]]: List of schema response
    """
    request_info = _get_schemas_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_schema_fields(id: str) -> List[Dict[str, Any]]:
    """
    Asynchronously get schema fields.

    Args:
        id (str): Schema ID

    Returns:
        List[Dict[str, Any]]: List of field response
    """
    request_info = _get_schema_fields_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_schema(request: SchemaRequest) -> Dict[str, Any]:
    """
    Asynchronously create schema.

    Args:
        request (SchemaRequest): Schema request

    Returns:
        Dict[str, Any]: Schema response
    """
    request_info = _create_schema_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_schema(id: str, request: SchemaRequest) -> Dict[str, Any]:
    """
    Asynchronously update schema.

    Args:
        id (str): Schema ID
        request (SchemaRequest): Schema request

    Returns:
        Dict[str, Any]: Schema response
    """
    request_info = _update_schema_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_schema_fields(id: str, request: List[FieldRequest]) -> List[Dict[str, Any]]:
    """
    Asynchronously update schema fields.

    Args:
        id (str): Schema ID
        request (List[FieldRequest]): List of field request

    Returns:
         List[Dict[str, Any]]: List of field response
    """
    request_info = _update_schema_fields_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_schema(id: str) -> None:
    """
    Asynchronously delete schema by ID.

    Args:
        id (str): Schema ID

    Returns:
        None
    """
    request_info = _delete_schema_request_info(id)
    return await __async_delete_request(**request_info)



from datetime import date
from typing import Optional

from algora.api.service.backtest.enum import BacktestType, BacktestStatus
from algora.common.base import Base


class BacktestRequest(Base):
    strategy_id: str
    type: BacktestType
    status: BacktestStatus
    start_date: date
    end_date: date
    portfolio: str  # TODO replace with portfolio object in Quant SDK


class BacktestResponse(Base):
    id: str
    strategy_id: str
    type: BacktestType
    status: BacktestStatus
    start_date: date
    end_date: date
    portfolio: str  # TODO replace with portfolio object in Quant SDK
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]


class BacktestViewResponse(Base):
    id: str
    strategy_id: str
    strategy_name: str
    type: BacktestType
    status: BacktestStatus
    start_date: date
    end_date: date
    portfolio: str  # TODO replace with portfolio object in Quant SDK
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]



from algora.api.service.research.runner.__util import _get_or_create_runner_request_info
from algora.common.decorators import data_request
from algora.common.requests import __put_request


@data_request(transformer=lambda data: data)
def get_or_create_runner() -> str:
    """
    Get or create runner.

    Returns:
        str: Runner ID
    """
    request_info = _get_or_create_runner_request_info()
    return __put_request(**request_info)



import json
from typing import List

from algora.api.service.datasets.field.model import FieldRequest
from algora.api.service.datasets.schema.model import SchemaRequest


def _get_schema_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/schema/{id}"
    }


def _get_schemas_request_info() -> dict:
    return {
        "endpoint": f"config/datasets/schema"
    }


def _get_schema_fields_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/schema/{id}/fields"
    }


def _create_schema_request_info(request: SchemaRequest) -> dict:
    return {
        "endpoint": f"config/datasets/schema",
        "json": request.request_dict()
    }


def _update_schema_request_info(id: str, request: SchemaRequest) -> dict:
    return {
        "endpoint": f"config/datasets/schema/{id}",
        "json": request.request_dict()
    }


def _update_schema_fields_request_info(id: str, request: List[FieldRequest]) -> dict:
    return {
        "endpoint": f"config/datasets/schema/{id}/fields",
        "json": [json.loads(f.json()) for f in request]
    }


def _delete_schema_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/schema/{id}"
    }



"""
Configuration classes used for configuring the Algora SDK.
"""
import os
from typing import Optional


class Auth:
    """
    Configuration class for getting the authentication information from the environment.
    """

    # Feature to keep values hidden from printing (not totally)
    # Allows for dynamic rendering of auth values
    @property
    def username(self) -> Optional[str]:
        return os.getenv("ALGORA_USER", None)

    @property
    def password(self) -> Optional[str]:
        return os.getenv("ALGORA_PWD", None)

    @property
    def access_token(self) -> Optional[str]:
        return os.getenv("ALGORA_ACCESS_TOKEN", None)

    @property
    def refresh_token(self) -> Optional[str]:
        return os.getenv("ALGORA_REFRESH_TOKEN", None)

    def can_authenticate(self) -> bool:
        """
        Check environment to see if the user provided enough info to authenticate requests.

        Returns:
            bool: A boolean that represents whether you can authenticate requests
        """
        auth_options = self.access_token or self.refresh_token or (self.username and self.password)
        return auth_options is not None

    def __eq__(self, other: "Auth"):
        return (
                self.access_token == other.access_token and
                self.username == self.username and
                self.password == self.password
        )


class EnvironmentConfig:
    """
    Configuration class for getting environment information.
    """
    urls: dict = {
        'algora': 'https://api.algoralabs.com',
        'keycloak_admin': 'https://auth.algoralabs.com/auth/admin/realms/production',
        'keycloak': 'https://auth.algoralabs.com/auth/realms/production',
        'fred': 'https://api.stlouisfed.org/fred'
    }
    auth_config: Auth = Auth()

    def get_url(self, key: str = "algora") -> str:
        """
        Get base URL given key.

        Args:
            key (str): URL key; 'algora', 'keycloak_admin', 'keycloak' or 'fred'

        Returns:
            str: Base URL string
        """
        return self.urls.get(key, self.urls["algora"])



"""
Asynchronous admin API requests.
"""
from typing import List, Dict

from algora.api.service.admin.__util import (
    _get_client_id_request_info, _get_users_request_info, _get_user_request_info, _update_user_request_info,
    _get_groups_request_info, _get_group_members_request_info, _add_group_to_user_request_info,
    _delete_group_from_user_request_info, _get_roles_request_info, _get_roles_for_user_request_info,
    _add_role_mapping_to_user_request_info, _execute_actions_email_request_info
)
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_get_request, __async_put_request, __async_post_request, __async_delete_request
)


@async_data_request(transformer=no_transform)
async def async_get_client_id():
    """
    Asynchronously get keycloak client ID.

    Returns:
        Keycloak client ID.
    """
    request_info = _get_client_id_request_info()
    return await __async_get_request(**request_info)


@async_data_request
async def async_get_users():
    """
    Asynchronously get all users.

    Returns:
        List of user response.
    """
    request_info = _get_users_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_user(user_id: str):
    """
    Asynchronously get user by ID.

    Args:
        user_id (str): User ID

    Returns:
        User response
    """
    request_info = _get_user_request_info(user_id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform, processor=lambda r: r)
def async_execute_actions_email(user_id: str, actions: List[str], client_id: str, redirect_uri: str):
    """
    Asynchronously update user.

    Args:
        user_id (str): User ID
        actions: (List[str]): Required user actions
        client_id: (str): Client ID
        redirect_uri: (str): Redirect URI after executing actions

    Returns:
        User response
    """
    request_info = _execute_actions_email_request_info(user_id, actions, client_id, redirect_uri)
    return __async_put_request(**request_info)


@async_data_request(transformer=no_transform, processor=lambda r: r)
async def async_update_user(user_id: str, data: dict):
    """
    Asynchronously update user.

    Args:
        user_id (str): User ID
        data (dict): User request

    Returns:
        User response
    """
    request_info = _update_user_request_info(user_id, data)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_groups():
    """
    Asynchronously get groups.

    Returns:
        List of groups.
    """
    request_info = _get_groups_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_group_members(id: str):
    """
    Asynchronously get group members given group ID.

    Args:
        id (str): Group ID

    Returns:
        List of user response
    """
    request_info = _get_group_members_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform, processor=lambda r: r)
async def async_add_group_to_user(user_id: str, group_id: str):
    """
    Asynchronously add user to group.

    Args:
        user_id (str): User ID
        group_id (str): Group ID

    Returns:
        None
    """
    request_info = _add_group_to_user_request_info(user_id, group_id)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform, processor=lambda r: r)
async def async_delete_group_from_user(user_id: str, group_id: str):
    """
    Asynchronously delete user from group.

    Args:
        user_id (str): User ID
        group_id (str): Group ID

    Returns:
        None
    """
    request_info = _delete_group_from_user_request_info(user_id, group_id)
    return await __async_delete_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_roles():
    """
    Asynchronously get roles.

    Returns:
        List of role response.
    """
    request_info = _get_roles_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_roles_for_user(user_id: str):
    """
    Asynchronously get roles for user.

    Args:
        user_id (str): User ID

    Returns:
        List of role response
    """
    request_info = _get_roles_for_user_request_info(user_id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform, processor=lambda r: r)
async def async_add_role_mapping_to_user(user_id: str, role_ids: List[Dict[str, str]]):
    """
    Asynchronously add role mapping to user.

    Args:
        user_id (str): User ID
        role_ids (List[Dict[str, str]]): List of role ID mapping

    Returns:
        None
    """
    request_info = _add_role_mapping_to_user_request_info(user_id, role_ids)
    return await __async_post_request(**request_info)



from typing import Dict, Any

import pandas as pd

from algora.api.service.backtest.state.__util import (
    _get_all_portfolio_state_request_info, _get_all_cash_payments_request_info, _create_portfolio_state_request_info,
    _create_cash_payment_request_info
)
from algora.api.service.backtest.state.model import PortfolioStateRequest, CashPaymentRequest
from algora.common.decorators import async_data_request
from algora.common.enum import Order
from algora.common.function import no_transform
from algora.common.requests import __async_put_request, __async_get_request


@async_data_request
async def async_get_all_portfolio_state(backtest_id: str, order: Order = Order.ASC, **kwargs) -> pd.DataFrame:
    """
    Asynchronously get all portfolio states.

    Args:
        backtest_id (str): Backtest ID
        order (Order): SQL Order
        **kwargs: Extra arguments

    Returns:
        DataFrame: DataFrame of portfolio states
    """
    request_info = _get_all_portfolio_state_request_info(backtest_id, order, **kwargs)
    return await __async_get_request(**request_info)


@async_data_request
async def async_get_all_cash_payments(backtest_id: str, order: Order = Order.ASC, **kwargs) -> pd.DataFrame:
    """
    Asynchronously get all cash payments.

    Args:
        backtest_id (str): Backtest ID
        order (Order): SQL Order
        **kwargs: Extra arguments

    Returns:
        DataFrame: DataFrame of cash payments
    """
    request_info = _get_all_cash_payments_request_info(backtest_id, order, **kwargs)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_portfolio_state(request: PortfolioStateRequest) -> Dict[str, Any]:
    """
    Asynchronously create portfolio state.

    Args:
        request (PortfolioStateRequest): Portfolio state request

    Returns:
        Dict[str, Any]: Portfolio state response
    """
    request_info = _create_portfolio_state_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_cash_payment(request: CashPaymentRequest) -> Dict[str, Any]:
    """
    Asynchronously create cash payment.

    Args:
        request (CashPaymentRequest): Cash payment request

    Returns:
        Dict[str, Any]: Cash payment response
    """
    request_info = _create_cash_payment_request_info(request)
    return await __async_put_request(**request_info)



"""
Helper functions.
"""
import calendar
from datetime import date, time, datetime
from functools import reduce
from typing import Dict, Any, Union, List, Optional

import pandas as pd
import pytz
from aiohttp import ClientResponse
from pandas import DataFrame, to_datetime, bdate_range
from pandas_market_calendars import get_calendar
from requests import Response


def coalesce(*args):
    """
    Coalesce operator returns the first arg that isn't None.

    Args:
        *args: Tuple of args passed to the function

    Returns:
        The first non-None value in *args
    """
    return reduce(lambda x, y: x if x is not None else y, args)


def coalesce_callables(*args):
    """
    Coalesce operator returns the first arg that isn't None. If the arg is callable it will check that the value
    returned from calling the arg is not none and then return the value from calling it.

    WARNING: If an argument implements __call__ this method will evaluate the return of the __call__ method and return
    that instead of the argument itself. This is important when using python classes.

    Args:
        *args: Tuple of args passed to the function

    Returns:
        The first non-None value in *args
    """
    for arg in args:
        value = arg() if callable(arg) else arg
        if value is not None:
            return value
    return None


def transform_one_or_many(
        data: Union[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]]],
        key: Optional[str] = None
) -> Union[DataFrame, Dict[str, DataFrame]]:
    """
    Converts data to a dataframe or multiple dataframes.

    Args:
        data (Union[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]]]): Dict being transformed into Dataframes
        key (Optional[str]): Key for indexing the values of the data dict

    Returns:
        Union[DataFrame, Dict[str, DataFrame]]: A single dataframe or a map of names to dataframes
    """
    if isinstance(data, dict):
        for k in data.keys():
            data[k] = DataFrame(data[k][key])
        return data

    return DataFrame(data)


def to_pandas_with_index(data: Dict[str, Any], index: str = 'date') -> DataFrame:
    """
    Transform input data into pandas dataframe and assign index.

    Args:
        data (Dict[str, Any]): Input data as dict
        index (str): Index column name

    Returns:
        DataFrame: DataFrame
    """
    # necessary to drop column in order to avoid duplicates when converting to json
    return pd.DataFrame(data).set_index(index, drop=True)


def no_transform(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Passthrough data without any transformation.

    Args:
        data (Dict[str, Any]): Input data

    Returns:
        Dict[str, Any]: Output data
    """
    return data


def timestamp_to_datetime(timestamp: int) -> datetime:
    """
    Transform epoch timestamp in milliseconds to datetime.

    Args:
        timestamp (int): Epoch timestamp in milliseconds

    Returns:
        datetime: Python datetime
    """
    return datetime.fromtimestamp(timestamp / 1000)


def date_to_timestamp(as_of_date: date) -> int:
    """
    Convert date to epoch timestamp in milliseconds.

    Args:
        as_of_date (date): Python date

    Returns:
        int: Epoch timestamp in milliseconds
    """
    return calendar.timegm(as_of_date.timetuple()) * 1000


def datetime_to_timestamp(dt: datetime) -> int:
    """
    Convert datetime to epoch timestamp in milliseconds.

    Args:
        dt (datetime): Python datetime

    Returns:
        int: Epoch timestamp in milliseconds
    """
    return calendar.timegm(dt.utctimetuple()) * 1000


def datetime_to_utc(dt: datetime) -> datetime:
    """
    Standardize datetime to UTC. Assume that datetime where `tzinfo=None` is already in UTC.

    Args:
        dt (datetime): Python datetime

    Returns:
        datetime: Python datetime with standardized UTC timezone (`tzinfo=None`)
    """
    # assume that datetime without timezone is already in UTC
    if dt.tzinfo is None:
        return dt
    return dt.astimezone(pytz.utc).replace(tzinfo=None)


def date_to_datetime(as_of_date: date, as_of_time: time = datetime.min.time()) -> datetime:
    """
    Convert date and optional time to datetime. NOTE: time should not contain a timezone or else offset may not be
    correct.

    Args:
        as_of_date (date): Python date
        as_of_time (time): Python time

    Returns:
        datetime: Python datetime
    """
    return datetime.combine(as_of_date, as_of_time)


def create_dates_between(start: date, end: date, frequency: str = 'B') -> List[date]:
    """
    Create dates between start and end date (inclusive). Frequency used to determine which days of the week are used.

    Args:
        start (date): Python date
        end (date): Python date
        frequency (str): Frequency for date range

    Returns:
        List[date]: List of dates
    """
    return [dt.date() for dt in to_datetime(bdate_range(start=start, end=end, freq=frequency).to_list())]


def create_market_dates_between(start: date, end: date, name: str = 'NYSE') -> List[date]:
    """
    Create dates between start and end date (inclusive) given exchange calendar.

    Args:
        start (date): Python date
        end (date): Python date
        name (str): Calendar name, such as 'NYSE'

    Returns:
        List[date]: List of dates
    """
    return [dt.date() for dt in
            to_datetime(get_calendar(name).schedule(start_date=start, end_date=end).index).to_list()]


async def build_response_obj(aio_response: ClientResponse) -> Response:
    """
    Construct a requests Response object from an aiohttp Client Response object

    Args:
        aio_response (ClientResponse): Response object from aiohttp

    Returns:
        Response: A requests Response object
    """
    content = await aio_response.content.read()

    response = Response()
    response._content = content
    response.url = str(aio_response.url)
    response.status_code = aio_response.status
    headers = {row[0]: row[1] for row in aio_response.headers.items()}
    response.headers = headers
    return response



"""
Common models.
"""
from algora.common.base import Base
from algora.common.enum import FieldType
from algora.common.type import Datetime


class Field(Base):
    logical_name: str
    type: FieldType


class DataRequest(Base):
    """
    Base data request, inherited by all instrument-specific data request classes.
    """
    date: Datetime



"""
Data API query functions for [IEX](https://iexcloud.io/) stock data.
"""
from algora.api.data.iex.stocks.asynchronous import (
    async_symbols, async_historical_prices, async_news, async_peer_group
)
from algora.api.data.iex.stocks.synchronous import symbols, historical_prices, news, peer_group



from algora.api.service.datasets.schema.asynchronous import (
    async_get_schema, async_get_schemas, async_get_schema_fields, async_create_schema, async_update_schema,
    async_update_schema_fields, async_delete_schema
)
from algora.api.service.datasets.schema.synchronous import (
    get_schema, get_schemas, get_schema_fields, create_schema, update_schema, update_schema_fields, delete_schema
)



from abc import abstractmethod, ABC

from algora.common.model import DataRequest
from algora.common.type import DataResponse


class Source(ABC):
    """
    Data source class, containing name and get_data methods.
    """

    @abstractmethod
    def name(self) -> str:
        """
        Data source name.

        Returns:
            str: Data source name
        """
        pass

    @abstractmethod
    def get_data(self, request: DataRequest) -> DataResponse:
        """
        Get data and return result.

        Args:
            request (DataRequest): Data request class

        Returns:
            DataResponse: Data response
        """
        pass


class AsyncSource(ABC):
    """
    Asynchronous data source class, containing name and asynchronous get_data methods.
    """

    @abstractmethod
    def name(self) -> str:
        """
        Data source name.

        Returns:
            str: Data source name
        """
        pass

    @abstractmethod
    async def get_data(self, request: DataRequest) -> DataResponse:
        """
        Asynchronously get data and return result.

        Args:
            request (DataRequest): Data request class

        Returns:
            DataResponse: Data response
        """
        pass



from algora.common.base import BaseEnum


class BacktestType(BaseEnum):
    STRATEGY = 'STRATEGY'
    BACKTEST = 'BACKTEST'


class BacktestStatus(BaseEnum):
    RUNNING = 'RUNNING'
    COMPLETED = 'COMPLETED'
    FAILED = 'FAILED'
    CANCELED = 'CANCELED'



"""
Common enums.
"""
from algora.common.base import BaseEnum


class Order(BaseEnum):
    ASC = 'ASC'
    DESC = 'DESC'


class FieldType(BaseEnum):
    BOOLEAN = 'BOOLEAN'
    DOUBLE = 'DOUBLE'
    INTEGER = 'INTEGER'
    TEXT = 'TEXT'
    TIMESTAMP = 'TIMESTAMP'
    DATETIME = 'DATETIME'
    UNKNOWN = 'UNKNOWN'


class PaymentType(BaseEnum):
    """Cash payment type, such as transaction cost"""
    MID_COST = 'MID_COST'
    SLIPPAGE = 'SLIPPAGE'


class Currency(BaseEnum):
    """Currency, ISO 4217 currency code or exchange quote modifier (e.g. GBP vs GBp)"""

    _ = ''
    ACU = 'ACU'
    ADP = 'ADP'
    AED = 'AED'
    AFA = 'AFA'
    ALL = 'ALL'
    AMD = 'AMD'
    ANG = 'ANG'
    AOA = 'AOA'
    AOK = 'AOK'
    AON = 'AON'
    ARA = 'ARA'
    ARS = 'ARS'
    ARZ = 'ARZ'
    ATS = 'ATS'
    AUD = 'AUD'
    AUZ = 'AUZ'
    AZM = 'AZM'
    AZN = 'AZN'
    B03 = 'B03'
    BAD = 'BAD'
    BAK = 'BAK'
    BAM = 'BAM'
    BBD = 'BBD'
    BDN = 'BDN'
    BDT = 'BDT'
    BEF = 'BEF'
    BGL = 'BGL'
    BGN = 'BGN'
    BHD = 'BHD'
    BIF = 'BIF'
    BMD = 'BMD'
    BND = 'BND'
    BOB = 'BOB'
    BR6 = 'BR6'
    BRE = 'BRE'
    BRF = 'BRF'
    BRL = 'BRL'
    BRR = 'BRR'
    BSD = 'BSD'
    BTC = 'BTC'
    BTN = 'BTN'
    BTR = 'BTR'
    BWP = 'BWP'
    BYR = 'BYR'
    BZD = 'BZD'
    C23 = 'C23'
    CAC = 'CAC'
    CAD = 'CAD'
    CAZ = 'CAZ'
    CCI = 'CCI'
    CDF = 'CDF'
    CFA = 'CFA'
    CHF = 'CHF'
    CHZ = 'CHZ'
    CLF = 'CLF'
    CLP = 'CLP'
    CLZ = 'CLZ'
    CNH = 'CNH'
    CNO = 'CNO'
    CNY = 'CNY'
    CNZ = 'CNZ'
    COP = 'COP'
    COZ = 'COZ'
    CPB = 'CPB'
    CPI = 'CPI'
    CRC = 'CRC'
    CUP = 'CUP'
    CVE = 'CVE'
    CYP = 'CYP'
    CZH = 'CZH'
    CZK = 'CZK'
    DAX = 'DAX'
    DEM = 'DEM'
    DIJ = 'DIJ'
    DJF = 'DJF'
    DKK = 'DKK'
    DOP = 'DOP'
    DZD = 'DZD'
    E51 = 'E51'
    E52 = 'E52'
    E53 = 'E53'
    E54 = 'E54'
    ECI = 'ECI'
    ECS = 'ECS'
    ECU = 'ECU'
    EEK = 'EEK'
    EF0 = 'EF0'
    EGP = 'EGP'
    ESP = 'ESP'
    ETB = 'ETB'
    EUR = 'EUR'
    EUZ = 'EUZ'
    F06 = 'F06'
    FED = 'FED'
    FIM = 'FIM'
    FJD = 'FJD'
    FKP = 'FKP'
    FRF = 'FRF'
    FT1 = 'FT1'
    GBP = 'GBP'
    GBZ = 'GBZ'
    GEK = 'GEK'
    GEL = 'GEL'
    GHC = 'GHC'
    GHS = 'GHS'
    GHY = 'GHY'
    GIP = 'GIP'
    GLD = 'GLD'
    GLR = 'GLR'
    GMD = 'GMD'
    GNF = 'GNF'
    GQE = 'GQE'
    GRD = 'GRD'
    GTQ = 'GTQ'
    GWP = 'GWP'
    GYD = 'GYD'
    HKB = 'HKB'
    HKD = 'HKD'
    HNL = 'HNL'
    HRK = 'HRK'
    HSI = 'HSI'
    HTG = 'HTG'
    HUF = 'HUF'
    IDB = 'IDB'
    IDO = 'IDO'
    IDR = 'IDR'
    IEP = 'IEP'
    IGP = 'IGP'
    ILS = 'ILS'
    INO = 'INO'
    INP = 'INP'
    INR = 'INR'
    IPA = 'IPA'
    IPX = 'IPX'
    IQD = 'IQD'
    IRR = 'IRR'
    IRS = 'IRS'
    ISI = 'ISI'
    ISK = 'ISK'
    ISO = 'ISO'
    ITL = 'ITL'
    J05 = 'J05'
    JMD = 'JMD'
    JNI = 'JNI'
    JOD = 'JOD'
    JPY = 'JPY'
    JPZ = 'JPZ'
    JZ9 = 'JZ9'
    KES = 'KES'
    KGS = 'KGS'
    KHR = 'KHR'
    KMF = 'KMF'
    KOR = 'KOR'
    KPW = 'KPW'
    KRW = 'KRW'
    KWD = 'KWD'
    KYD = 'KYD'
    KZT = 'KZT'
    LAK = 'LAK'
    LBA = 'LBA'
    LBP = 'LBP'
    LHY = 'LHY'
    LKR = 'LKR'
    LRD = 'LRD'
    LSL = 'LSL'
    LSM = 'LSM'
    LTL = 'LTL'
    LUF = 'LUF'
    LVL = 'LVL'
    LYD = 'LYD'
    MAD = 'MAD'
    MDL = 'MDL'
    MGF = 'MGF'
    MKD = 'MKD'
    MMK = 'MMK'
    MNT = 'MNT'
    MOP = 'MOP'
    MRO = 'MRO'
    MTP = 'MTP'
    MUR = 'MUR'
    MVR = 'MVR'
    MWK = 'MWK'
    MXB = 'MXB'
    MXN = 'MXN'
    MXP = 'MXP'
    MXW = 'MXW'
    MXZ = 'MXZ'
    MYO = 'MYO'
    MYR = 'MYR'
    MZM = 'MZM'
    MZN = 'MZN'
    NAD = 'NAD'
    ND3 = 'ND3'
    NGF = 'NGF'
    NGI = 'NGI'
    NGN = 'NGN'
    NIC = 'NIC'
    NLG = 'NLG'
    NOK = 'NOK'
    NOZ = 'NOZ'
    NPR = 'NPR'
    NZD = 'NZD'
    NZZ = 'NZZ'
    O08 = 'O08'
    OMR = 'OMR'
    PAB = 'PAB'
    PEI = 'PEI'
    PEN = 'PEN'
    PEZ = 'PEZ'
    PGK = 'PGK'
    PHP = 'PHP'
    PKR = 'PKR'
    PLN = 'PLN'
    PLZ = 'PLZ'
    PSI = 'PSI'
    PTE = 'PTE'
    PYG = 'PYG'
    QAR = 'QAR'
    R2K = 'R2K'
    ROL = 'ROL'
    RON = 'RON'
    RSD = 'RSD'
    RUB = 'RUB'
    RUF = 'RUF'
    RUR = 'RUR'
    RWF = 'RWF'
    SAR = 'SAR'
    SBD = 'SBD'
    SCR = 'SCR'
    SDP = 'SDP'
    SDR = 'SDR'
    SEK = 'SEK'
    SET = 'SET'
    SGD = 'SGD'
    SGS = 'SGS'
    SHP = 'SHP'
    SKK = 'SKK'
    SLL = 'SLL'
    SRG = 'SRG'
    SSI = 'SSI'
    STD = 'STD'
    SUR = 'SUR'
    SVC = 'SVC'
    SVT = 'SVT'
    SYP = 'SYP'
    SZL = 'SZL'
    T21 = 'T21'
    T51 = 'T51'
    T52 = 'T52'
    T53 = 'T53'
    T54 = 'T54'
    T55 = 'T55'
    T71 = 'T71'
    TE0 = 'TE0'
    TED = 'TED'
    TF9 = 'TF9'
    THB = 'THB'
    THO = 'THO'
    TMM = 'TMM'
    TND = 'TND'
    TNT = 'TNT'
    TOP = 'TOP'
    TPE = 'TPE'
    TPX = 'TPX'
    TRB = 'TRB'
    TRL = 'TRL'
    TRY = 'TRY'
    TRZ = 'TRZ'
    TTD = 'TTD'
    TWD = 'TWD'
    TZS = 'TZS'
    UAH = 'UAH'
    UCB = 'UCB'
    UDI = 'UDI'
    UFC = 'UFC'
    UFZ = 'UFZ'
    UGS = 'UGS'
    UGX = 'UGX'
    USB = 'USB'
    USD = 'USD'
    UVR = 'UVR'
    UYP = 'UYP'
    UYU = 'UYU'
    UZS = 'UZS'
    VAC = 'VAC'
    VEB = 'VEB'
    VEF = 'VEF'
    VES = 'VES'
    VND = 'VND'
    VUV = 'VUV'
    WST = 'WST'
    XAF = 'XAF'
    XAG = 'XAG'
    XAU = 'XAU'
    XPD = 'XPD'
    XPT = 'XPT'
    XCD = 'XCD'
    XDR = 'XDR'
    XEU = 'XEU'
    XOF = 'XOF'
    XPF = 'XPF'
    YDD = 'YDD'
    YER = 'YER'
    YUD = 'YUD'
    YUN = 'YUN'
    ZAL = 'ZAL'
    ZAR = 'ZAR'
    ZAZ = 'ZAZ'
    ZMK = 'ZMK'
    ZMW = 'ZMW'
    ZRN = 'ZRN'
    ZRZ = 'ZRZ'
    ZWD = 'ZWD'
    AUd = 'AUd'
    BWp = 'BWp'
    EUr = 'EUr'
    GBp = 'GBp'
    ILs = 'ILs'
    KWd = 'KWd'
    MWk = 'MWk'
    SGd = 'SGd'
    SZl = 'SZl'
    USd = 'USd'
    ZAr = 'ZAr'



from typing import Dict, Any, List

from algora.api.service.backtest.__util import (
    _get_backtest_request_info, _get_backtests_request_info, _create_backtest_request_info,
    _update_backtest_request_info, _delete_backtest_request_info
)
from algora.api.service.backtest.enum import BacktestStatus
from algora.api.service.backtest.model import BacktestRequest
from algora.common.decorators import data_request
from algora.common.function import no_transform
from algora.common.requests import __get_request, __put_request, __post_request, __delete_request


@data_request(transformer=no_transform)
def get_backtest(id: str) -> Dict[str, Any]:
    """
    Get backtest by ID.

    Args:
        id (str): Backtest ID

    Returns:
        Dict[str, Any]: Backtest response
    """
    request_info = _get_backtest_request_info(id)
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def get_backtests() -> List[Dict[str, Any]]:
    """
    Get all backtests.

    Returns:
        List[Dict[str, Any]]: List of backtest response
    """
    request_info = _get_backtests_request_info()
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def create_backtest(request: BacktestRequest) -> Dict[str, Any]:
    """
    Create backtest.

    Args:
        request (BacktestRequest): Backtest request

    Returns:
        Dict[str, Any]: Backtest response
    """
    request_info = _create_backtest_request_info(request)
    return __put_request(**request_info)


@data_request(transformer=no_transform)
def update_backtest(id: str, status: BacktestStatus) -> Dict[str, Any]:
    """
    Update backtest.

    Args:
        id (str): Backtest ID
        status (BacktestRequest): Backtest request

    Returns:
        Dict[str, Any]: Backtest response
    """
    request_info = _update_backtest_request_info(id, status)
    return __post_request(**request_info)


@data_request(transformer=no_transform)
def delete_backtest(id: str) -> None:
    """
    Delete backtest by ID.

    Args:
        id (str): Backtest ID

    Returns:
        None
    """
    request_info = _delete_backtest_request_info(id)
    return __delete_request(**request_info)



import json
import logging
from typing import Optional

from requests import Response

from algora.common.errors import AuthenticationError

logger = logging.getLogger(__name__)


def _handle_auth_response(auth_response: Response, error_msg: str) -> dict:
    if auth_response.status_code == 200:
        bearer_token = auth_response.json()['access_token']
        return {'Authorization': f'Bearer {bearer_token}'}
    else:
        error = AuthenticationError(error_msg)
        logger.error(error)
        raise error


def _access_token(access_token: str) -> dict:
    return json.loads(access_token)


def _validate_headers(**kwargs):
    headers = kwargs.get("headers", {})

    if headers.get("Authorization") is None:
        error = AuthenticationError("Authentication for the package was configured incorrectly and is either "
                                    "missing a ALGORA_ACCESS_TOKEN or ALGORA_REFRESH_TOKEN or ALGORA_USER and "
                                    "ALGORA_PWD environment variable(s)")
        logger.error(error)
        # TODO this fails when using FRED API (outside of Algora API, so it doesn't need Authorization header)
        #  raise error



import functools
from typing import Callable, Any

import numpy as np


def vectorize(function: Callable = None):
    """
    Decorator for wrapping numpy function in `np.vectorize(...)`

    Args:
        function (Callable): Numpy function

    Returns:
        Callable: Wrapped function
    """

    @functools.wraps(vectorize)
    def decorator(f):
        @functools.wraps(f)
        def wrap(*args, **kwargs) -> Any:
            vectorized = np.vectorize(f)
            return vectorized(*args, **kwargs)

        return wrap

    if function is None:
        return decorator
    return decorator(function)



from pandas import DataFrame

from algora.api.data.fred.__util import _transform_fred_observations, _get_series_info
from algora.api.data.fred.model import FredQuery
from algora.common.decorators import data_request
from algora.common.requests import __get_request


@data_request(transformer=_transform_fred_observations)
def get_series(query: FredQuery) -> DataFrame:
    """
    Get FRED series.

    Args:
        query (FredQuery): FRED query

    Returns:
        DataFrame: DataFrame
    """
    request_info = _get_series_info(query)
    return __get_request(**request_info)



from typing import Dict, Any

import pandas as pd
from pandas import DataFrame

from algora.api.data.fred.model import FredQuery


def _transform_fred_observations(data: Dict[str, Any]) -> DataFrame:
    return pd.DataFrame(data['observations'])


def _get_series_info(query: FredQuery) -> dict:
    return {
        "endpoint": "/series/observations",
        "url_key": "fred",
        "params": query.dict(exclude_none=True)
    }



from typing import Optional, List

from algora.api.service.datasets.enum import DatasetDataType
from algora.common.base import Base


class DatasetRequest(Base):
    display_name: str
    logical_name: str
    description: Optional[str]
    data_type: DatasetDataType
    data_query: str
    data_query_type: str
    schema_id: str
    directory_id: str


class DatasetSearchRequest(Base):
    query: str
    data_types: Optional[List[str]]  # TODO make enum



"""
Document API.
"""
from algora.api.service.documents.asynchronous import (
    async_get_document, async_search_documents, async_create_document, async_update_document, async_delete_document
)
from algora.api.service.documents.synchronous import (
    get_document, search_documents, create_document, update_document, delete_document
)



def _get_resource_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/resource/{id}/resource"
    }


def _delete_resource_request_info(id: str):
    return {
        'endpoint': f"config/resource/{id}"
    }



"""
Admin API.
"""
from algora.api.service.admin.asynchronous import (
    async_get_client_id, async_get_users, async_get_user, async_update_user, async_get_groups, async_get_group_members,
    async_add_group_to_user, async_delete_group_from_user, async_get_roles, async_get_roles_for_user,
    async_add_role_mapping_to_user
)
from algora.api.service.admin.synchronous import (
    get_client_id, get_users, get_user, update_user, get_groups, get_group_members, add_group_to_user,
    delete_group_from_user, get_roles, get_roles_for_user, add_role_mapping_to_user
)



"""
Data engine API (deprecated).
"""
from algora.api.service.data_engine.asynchronous import async_analyze_data, async_transform_data
from algora.api.service.data_engine.synchronous import analyze_data, transform_data



from typing import Union, Dict, List

from pandas import DataFrame

from algora.api.data.iex.__util import __base_request
from algora.api.data.iex.stocks.__util import (
    _symbols_request_info, _historical_prices_request_info, _news_request_info, _peer_group_request_info
)
from algora.common.decorators import data_request
from algora.common.function import transform_one_or_many, no_transform


@data_request
def symbols() -> DataFrame:
    """
    Wrapper for [IEX symbols API](https://iexcloud.io/docs/api/#symbols) to get supported symbols.

    Returns:
         DataFrame: IEX supported symbols
    """
    request_info = _symbols_request_info()
    return __base_request(**request_info)


@data_request(transformer=lambda d: transform_one_or_many(d, 'chart'))
def historical_prices(*symbol: str, **kwargs) -> Union[DataFrame, Dict[str, DataFrame]]:
    """
    Wrapper for IEX historical stock prices via
    [Time Series API](https://iexcloud.io/docs/api/#time-series-endpoint).

    Args:
         symbol (*str): Stock symbol(s), such as "AAPL" or "AAPL", "FB"
         **kwargs: Optional args to pass to the IEX API

    Returns:
        Union[DataFrame, Dict[str, DataFrame]]: Historical prices for the symbol(s) requested
    """
    # default query params
    request_info = _historical_prices_request_info(*symbol, **kwargs)
    return __base_request(**request_info)


@data_request
def news(symbol: str, **kwargs) -> DataFrame:
    """
    Wrapper for [IEX stock news for given symbol](https://iexcloud.io/docs/api/#news).

    Args:
        symbol (str): Stock symbol, such as AAPL
        **kwargs: Optional args to pass to the IEX API

    Returns:
        DataFrame: News for the symbol requested
    """
    request_info = _news_request_info(symbol, kwargs)
    return __base_request(**request_info)


@data_request(transformer=no_transform)
def peer_group(symbol: str) -> List[str]:
    """
    Wrapper for [IEX stock peer group](https://iexcloud.io/docs/api/#peers).

    Args:
        symbol (str): Stock symbol, such as AAPL

    Returns:
         List[str]: List of symbols
    """
    request_info = _peer_group_request_info(symbol)
    return __base_request(**request_info)



from typing import Dict, Any, List

from algora.api.service.datasets.field_group.__util import (
    _get_field_group_request_info, _get_field_groups_request_info, _create_field_group_request_info,
    _update_field_group_request_info, delete_field_group_request_info
)
from algora.api.service.datasets.field_group.model import FieldGroupRequest
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_get_request, __async_put_request, __async_post_request, __async_delete_request
)


@async_data_request(transformer=no_transform)
async def async_get_field_group(id: str) -> Dict[str, Any]:
    """
    Asynchronously get field group by ID.

    Args:
        id (str): Field group ID

    Returns:
        Dict[str, Any]: Field group response
    """
    request_info = _get_field_group_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_field_groups() -> List[Dict[str, Any]]:
    """
    Asynchronously get all field groups.

    Returns:
        List[Dict[str, Any]]: List of field group response
    """
    request_info = _get_field_groups_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_field_group(request: FieldGroupRequest) -> Dict[str, Any]:
    """
    Asynchronously create field group.

    Args:
        request (FieldGroupRequest): Field group request

    Returns:
        Dict[str, Any]: Field group response
    """
    request_info = _create_field_group_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_field_group(id: str, request: FieldGroupRequest) -> Dict[str, Any]:
    """
    Asynchronously update field group.

    Args:
        id (str): Field group ID
        request (FieldGroupRequest): Field group request

    Returns:
        Dict[str, Any]: Field group response
    """
    request_info = _update_field_group_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_field_group(id: str) -> None:
    """
    Asynchronously delete field group by ID.

    Args:
        id (str): Field group ID

    Returns:
        None
    """
    request_info = delete_field_group_request_info(id)
    return await __async_delete_request(**request_info)



from typing import Dict, Any

from algora.api.service.research.__util import _get_research_request_info
from algora.common.decorators.data import async_data_request
from algora.common.requests import __async_get_request


@async_data_request(transformer=lambda data: data)
async def async_get_research(id: str) -> Dict[str, Any]:
    """
    Asynchronously get research by ID.

    Args:
        id (str): Research ID

    Returns:
        Dict[str, Any]: Research response
    """
    request_info = _get_research_request_info(id)
    return await __async_get_request(**request_info)



"""
Base asynchronous and synchronous request functions inherited by all API request functions.
"""
from algora.common.requests.asynchronous import (
    __async_get_request, __async_put_request, __async_post_request, __async_delete_request
)
from algora.common.requests.synchronous import __get_request, __put_request, __post_request, __delete_request



from typing import Optional

import pandas as pd

from algora.api.service.data_engine.model import TransformOverride


def _analyze_data_request_info(data: pd.DataFrame) -> dict:
    parquet_bytes = data.to_parquet()

    return {
        'endpoint': "data-engine/alpha/analyze",
        'files': {'file': parquet_bytes}
    }


def _transform_data_request_info(data: pd.DataFrame, transform_override: Optional[TransformOverride] = None) -> dict:
    if transform_override is not None:
        transform_override = {'transform_override': transform_override.json()}
    else:
        transform_override = None

    parquet_bytes = data.to_parquet()

    return {
        'endpoint': f"data-engine/alpha/transform",
        'data': transform_override,
        'files': {'file': parquet_bytes}
    }



from pandas import DataFrame

from algora.api.data.fred.__util import _transform_fred_observations, _get_series_info
from algora.api.data.fred.model import FredQuery
from algora.common.decorators import async_data_request
from algora.common.requests import __async_get_request


@async_data_request(transformer=_transform_fred_observations)
async def async_get_series(query: FredQuery):
    """
    Asynchronously get FRED series.

    Args:
        query (FredQuery): FRED query

    Returns:
        DataFrame: DataFrame
    """
    request_info = _get_series_info(query)
    return await __async_get_request(**request_info)



import functools
from typing import Optional, Callable, Tuple, Dict, Any

from typing_extensions import Literal

from algora.common.data_source import Source, AsyncSource
from algora.common.model import DataRequest


def source(get_data_method: Callable[[DataRequest], Any] = None,
           *,
           name: Optional[str] = None,
           source_type: Literal['sync', 'async'] = 'sync'):
    """
    Data source function, containing name and asynchronous get_data functionality. Alternative to creating a Source
    class.

    Args:
        get_data_method (Callable[[DataRequest], Any]): Data request-response function to wrap
        name (Optional[str]): Data source name
        source_type (Literal['sync', 'async']): Type of source, either synchronous or asynchronous

    Returns:
        An synchronous or asynchronous source class containing the decorated method
    """

    @functools.wraps(get_data_method)
    def decorator(f):
        @functools.wraps(f)
        def wrap(*args: Tuple, **kwargs: Dict[str, Any]):
            """
            Wrapper for the decorated function.

            Args:
                *args: args for the function
                **kwargs: keyword args for the function

            Returns:

            """
            source_cls = AsyncSource if source_type == 'async' else Source
            cls = type(name, (source_cls,), {})
            setattr(cls, "get_data", f)
            setattr(cls, "name", lambda: name)
            return cls

        return wrap

    if get_data_method is None:
        return decorator
    return decorator(get_data_method)



import subprocess
import sys

import setuptools

try:
    import pypandoc

    long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()


def get_tag():
    tag = subprocess.getoutput('git tag --sort=version:refname | tail -n1')
    commits = subprocess.getoutput(f'git rev-list {tag}..HEAD --count')
    return f'{tag}.{commits}'


def is_public():
    if '--public' in sys.argv:
        sys.argv.remove('--public')
        return True
    return False


def get_packages_arguments():
    default_args = {
        'include': ['algora*'],
        'exclude': ['tests.*']
    }

    if is_public():
        default_args.update({
            'exclude': ['tests.*', '*quant*']
        })

    return default_args


setuptools.setup(
    name="algora-sdk",
    version=get_tag(),
    author="Algora Labs",
    author_email="hello@algoralabs.com",
    description="Algora Labs Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://docs.algoralabs.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.7",
    packages=setuptools.find_packages(**get_packages_arguments()),
    install_requires=[
        "requests==2.28.1",
        "numpy==1.22.1",
        "pandas==1.3.5",
        "cachetools==5.2.0",
        "pydash==5.1.0",
        "PyYaml==6.0",
        "pydantic==1.9.1",
        "asyncio==3.4.3",
        "aiohttp==3.8.1",
        "aiocache==0.11.1",
        "ujson==5.4.0",
        "fastparquet==0.8.1",
        "scipy==1.7.3",
        "pandas-market-calendars==3.4"
    ]
)



from typing import List, Dict


def _get_client_id_request_info() -> dict:
    return {
        "endpoint": "clients?clientId=realm-management",
        "url_key": "keycloak_admin"
    }


def _get_users_request_info() -> dict:
    return {
        "endpoint": "users",
        "params": {"max": 500},
        "url_key": "keycloak_admin"
    }


def _get_user_request_info(user_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}",
        "url_key": "keycloak_admin"
    }


def _execute_actions_email_request_info(user_id: str, actions: List[str], client_id: str, redirect_uri: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/execute-actions-email",
        "params": {"client_id": client_id, "redirect_uri": redirect_uri, "lifespan": 604800},
        "url_key": "keycloak_admin",
        "json": actions
    }


def _update_user_request_info(user_id: str, data: dict) -> dict:
    return {
        "endpoint": f"users/{user_id}",
        "url_key": "keycloak_admin",
        "json": data
    }


def _get_groups_request_info() -> dict:
    return {
        "endpoint": "groups",
        "url_key": "keycloak_admin"
    }


def _get_group_members_request_info(id: str) -> dict:
    return {
        "endpoint": f"groups/{id}/members",
        "url_key": "keycloak_admin"
    }


def _add_group_to_user_request_info(user_id: str, group_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/groups/{group_id}",
        "url_key": "keycloak_admin"
    }


def _delete_group_from_user_request_info(user_id: str, group_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/groups/{group_id}",
        "url_key": "keycloak_admin"
    }


def _get_roles_request_info() -> dict:
    return {
        "endpoint": "roles",
        "url_key": "keycloak_admin"
    }


def _get_roles_for_user_request_info(user_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/role-mappings",
        "url_key": "keycloak_admin"
    }


def _add_role_mapping_to_user_request_info(user_id: str, role_ids: List[Dict[str, str]]) -> dict:
    return {
        "endpoint": f"users/{user_id}/role-mappings/realm",
        "url_key": "keycloak_admin",
        "json": role_ids
    }



from algora.api.service.permission.model import PermissionRequest


def _get_permission_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/permission/{id}"
    }


def _get_permission_by_resource_id_request_info(resource_id: str) -> dict:
    return {
        'endpoint': f"config/permission/resource/{resource_id}"
    }


def _get_permissions_by_resource_id_request_info(resource_id: str) -> dict:
    return {
        'endpoint': f"config/permission/resource/{resource_id}/permissions"
    }


def _create_permission_request_info(request: PermissionRequest) -> dict:
    return {
        'endpoint': f"config/permission",
        'json': request.request_dict()
    }


def _update_permission_request_info(id: str, request: PermissionRequest) -> dict:
    return {
        'endpoint': f"config/permission/{id}",
        'json': request.request_dict()
    }


def _delete_permission_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/permission/{id}"
    }



from typing import Optional, Union, Any, Dict, List

from algora.api.data.sdr.enum import Repository, AssetClass
from algora.api.data.sdr.model import APIFieldFilter, FieldFilter, LogicalDisplayName, DataFilter


def __transform_filter(field_filter: Union[FieldFilter, APIFieldFilter]) -> APIFieldFilter:
    if isinstance(field_filter, APIFieldFilter):
        return field_filter

    return APIFieldFilter(
        logical_display=LogicalDisplayName(
            logical_name=field_filter.field,
            display_name=field_filter.field
        ),
        operator=field_filter.operator,
        selected_values=field_filter.selected_values
    )


def transform_data_filter(data_filter: Optional[DataFilter]) -> Optional[Dict[str, Any]]:
    """
    Transforms a data filter to dict.

    Args:
        data_filter (Optional[DataFilter]): Data filter to convert to a dict

    Returns:
        Optional[Dict[str, Any]]: A dict representation of a data filter
    """
    if data_filter is not None:
        transformed_filter = DataFilter(
            date_range=data_filter.date_range,
            filters=[__transform_filter(f) for f in data_filter.filters]
        )

        return transformed_filter.dict()

    return None


def _get_by_date_request_info(asset_class: AssetClass, date: str, repos: Optional[List[Repository]] = None) -> dict:
    if repos is None:
        repos = [Repository.CME, Repository.DTCC, Repository.ICE]

    repos_param = ",".join([repo.name for repo in repos])
    return {
        'endpoint': f"data/sdr/{asset_class.value}/{date}?repository={repos_param}"
    }


def _get_distinct_in_field_request_info(asset_class: AssetClass, field: str) -> dict:
    return {
        'endpoint': f"data/sdr/{asset_class.value}/{field}/distinct"
    }


def _commodity_request_info(filter: Optional[DataFilter] = None) -> dict:
    return {
        'id': "2880e242-8db4-49e2-aad3-e0339931582e",
        'json': transform_data_filter(filter)
    }


def _credit_request_info(filter: Optional[DataFilter] = None) -> dict:
    return {
        'id': "04863ce6-b179-420c-bef4-eb71f5391141",
        'json': transform_data_filter(filter)
    }


def _equity_request_info(filter: Optional[DataFilter] = None) -> dict:
    return {
        'id': "0f839686-a878-473b-a8a9-d2de2dcdd42c",
        'json': transform_data_filter(filter)
    }


def _forex_request_info(filter: Optional[DataFilter] = None) -> dict:
    return {
        'id': "f1137a7c-13db-451b-9603-f17dfa8bb147",
        'json': transform_data_filter(filter)
    }


def _rates_request_info(filter: Optional[DataFilter] = None) -> dict:
    return {
        'id': "a812f19c-354c-48e9-b86e-af055c631fcc",
        'json': transform_data_filter(filter)
    }



from algora.api.service.documents.model import DocumentRequest, SearchDocumentRequest


def _get_document_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/documents/{id}"
    }


def _search_documents_request_info(request: SearchDocumentRequest) -> dict:
    return {
        'endpoint': "config/documents/search",
        'json': request.request_dict()
    }


def _create_document_request_info(request: DocumentRequest) -> dict:
    return {
        'endpoint': "config/documents",
        'json': request.request_dict()
    }


def _update_document_request_info(id: str, request: DocumentRequest) -> dict:
    return {
        'endpoint': f"config/documents/{id}",
        'json': request.request_dict()
    }


def _delete_document_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/documents/{id}"
    }



from algora.api.service.datasets.model import DatasetSearchRequest, DatasetRequest


def _get_dataset_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/dataset/{id}"
    }


def _get_datasets_request_info() -> dict:
    return {
        "endpoint": f"config/datasets/dataset"
    }


def _search_datasets_request_info(request: DatasetSearchRequest) -> dict:
    return {
        "endpoint": f"config/datasets/dataset/search",
        "json": request.request_dict()
    }


def _create_dataset_request_info(request: DatasetRequest) -> dict:
    return {
        "endpoint": f"config/datasets/dataset",
        "json": request.request_dict()
    }


def _update_dataset_request_info(id: str, request: DatasetRequest) -> dict:
    return {
        "endpoint": f"config/datasets/dataset/{id}",
        "json": request.request_dict()
    }


def _delete_dataset_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/dataset/{id}"
    }



"""
Base classes with custom attributes for updating, serializing and deserializing data classes and enums.
"""
import json
from abc import ABC
from datetime import date, datetime, tzinfo
from enum import Enum

from pydantic import BaseModel, create_model

from algora.common.function import date_to_timestamp, datetime_to_timestamp


class BaseEnum(str, Enum):
    """
    Base enum class used by all enum classes.

    Note: Inheriting from str is necessary to correctly serialize output of enum
    """
    pass


class Base(ABC, BaseModel):
    """
    Base class used for all data classes.
    """

    class Config:
        # use enum values when using .dict() on object
        use_enum_values = True

        json_encoders = {
            date: date_to_timestamp,
            datetime: datetime_to_timestamp,
            tzinfo: str
        }

    @classmethod
    def cls_name(cls) -> str:
        """
        Get class name.

        Returns:
            str: Class name
        """
        return cls.__name__

    @classmethod
    def new_fields(cls, *args, **kwargs):
        """
        TODO DOCUMENT @will

        Args:
            *args:
            **kwargs:

        Returns:

        """
        return {}

    @classmethod
    def update(cls, *args, **kwargs):
        """
        TODO DOCUMENT @will

        Args:
            *args:
            **kwargs:

        Returns:

        """
        new_fields = cls.new_fields(*args, **kwargs)
        return create_model(cls.__name__, __base__=cls, **new_fields)

    def request_dict(self) -> dict:
        """
        Convert data class to dict. Used instead of `.dict()` to serialize dates as timestamps.

        Returns:
            dict: Serialized data class as dict
        """
        return json.loads(self.json())



from typing import Dict, Any

from algora.api.service.datasets.field.__util import (
    _delete_field_request_info, _create_field_request_info, _get_field_request_info, _update_field_request_info
)
from algora.api.service.datasets.field.model import FieldRequest
from algora.common.decorators import data_request
from algora.common.function import no_transform
from algora.common.requests import __delete_request, __post_request, __put_request, __get_request


@data_request(transformer=no_transform)
def get_field(id: str) -> Dict[str, Any]:
    """
    Get field by ID.

    Args:
        id (str): Field ID

    Returns:
        Dict[str, Any]: Field response
    """
    request_info = _get_field_request_info(id)
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def create_field(request: FieldRequest) -> Dict[str, Any]:
    """
    Create field.

    Args:
        request (FieldRequest): Field request

    Returns:
        Dict[str, Any]: Field response
    """
    request_info = _create_field_request_info(request)
    return __put_request(**request_info)


@data_request(transformer=no_transform)
def update_field(id: str, request: FieldRequest) -> Dict[str, Any]:
    """
    Update field.

    Args:
        id (str): Field ID
        request (FieldRequest): Field request

    Returns:
        Dict[str, Any]: Field response
    """
    request_info = _update_field_request_info(id, request)
    return __post_request(**request_info)


@data_request(transformer=no_transform)
def delete_field(id: str) -> None:
    """
    Delete field by ID.

    Args:
        id (str): Field ID

    Returns:
        None
    """
    request_info = _delete_field_request_info(id, )
    return __delete_request(**request_info)



"""
IEX API requests.
"""
from algora.common.requests import __get_request, __async_get_request


def __build_url(extension: str) -> str:
    return f"data/datasets/query/iex/{extension}"


def __base_request(extension: str, **kwargs):
    """
    Base request util for IEX API calls

    Args:
        extension (str): URI extension
        **kwargs: Request query params

    Returns:
        Response: HTTP response object
    """
    return __get_request(endpoint=__build_url(extension), params=kwargs)


async def __async_base_request(extension: str, **kwargs):
    """
    Asynchronous base request util for IEX API calls

    Args:
        extension (str): URI extension
        **kwargs: Request query params

    Returns:
        Response: HTTP response object
    """
    return await __async_get_request(endpoint=__build_url(extension), params=kwargs)



from typing import Dict, Any, List

from algora.api.service.datasets.__util import (
    _get_dataset_request_info, _get_datasets_request_info, _search_datasets_request_info, _create_dataset_request_info,
    _update_dataset_request_info, _delete_dataset_request_info
)
from algora.api.service.datasets.model import DatasetSearchRequest, DatasetRequest
from algora.common.decorators import data_request
from algora.common.function import no_transform
from algora.common.requests import __get_request, __post_request, __put_request, __delete_request


@data_request(transformer=no_transform)
def get_dataset(id: str) -> Dict[str, Any]:
    """
    Get dataset by ID.

    Args:
        id (str): Dataset ID

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _get_dataset_request_info(id)
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def get_datasets() -> List[Dict[str, Any]]:
    """
    Get all datasets.

    Returns:
        List[Dict[str, Any]]: List of dataset response
    """
    request_info = _get_datasets_request_info()
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def search_datasets(request: DatasetSearchRequest) -> List[Dict[str, Any]]:
    """
    Search all datasets.

    Args:
        request (DatasetSearchRequest): Dataset search request

    Returns:
        List[Dict[str, Any]]: List of dataset response
    """
    request_info = _search_datasets_request_info(request)
    return __post_request(**request_info)


@data_request(transformer=no_transform)
def create_dataset(request: DatasetRequest) -> Dict[str, Any]:
    """
    Create dataset.

    Args:
        request (DatasetRequest): Dataset request

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _create_dataset_request_info(request)
    return __put_request(**request_info)


@data_request(transformer=no_transform)
def update_dataset(id: str, request: DatasetRequest) -> Dict[str, Any]:
    """
    Update dataset.

    Args:
        id (str): Dataset ID
        request (DatasetRequest): Dataset request

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _update_dataset_request_info(id, request)
    return __post_request(**request_info)


@data_request(transformer=no_transform)
def delete_dataset(id: str) -> None:
    """
    Delete dataset by ID.

    Args:
        id (str): Dataset ID

    Returns:
        None
    """
    request_info = _delete_dataset_request_info(id)
    return __delete_request(**request_info)



"""
Common decorators.
"""
from algora.common.decorators.authorization import authenticated_request, async_authenticated_request
from algora.common.decorators.data import data_request, async_data_request
from algora.common.decorators.serializable import serializable
from algora.common.decorators.source import source



"""
Dataset API.
"""
from algora.api.service.datasets.asynchronous import (
    async_get_dataset, async_get_datasets, async_search_datasets, async_create_dataset, async_update_dataset,
    async_delete_dataset
)
from algora.api.service.datasets.synchronous import (
    get_dataset, get_datasets, search_datasets, create_dataset, update_dataset, delete_dataset
)



from typing import Optional, List

from pandas import DataFrame

from algora.api.data.query import query_dataset
from algora.api.data.sdr.__util import (
    _get_by_date_request_info, _get_distinct_in_field_request_info, _commodity_request_info, _credit_request_info,
    _rates_request_info, _forex_request_info, _equity_request_info
)
from algora.api.data.sdr.enum import AssetClass, Repository
from algora.api.data.sdr.model import DataFilter
from algora.common.decorators import data_request
from algora.common.function import to_pandas_with_index, no_transform
from algora.common.requests import __get_request


@data_request(transformer=lambda data: to_pandas_with_index(data, index='execution_timestamp'))
def get_by_date(asset_class: AssetClass, date: str, repos: Optional[List[Repository]] = None) -> DataFrame:
    """
    Get all SDR data by asset class, date and repositories.

    Args:
        asset_class (AssetClass): Asset class enum (COMMODITY, CREDIT, EQUITY, FOREX, RATES)
        date (str): Date in YYYY-MM-DD format (e.g. "2022-01-01")
        repos (Optional[List[Repository]]): Repository enum list (e.g. [CME, DTCC, ICE])

    Returns:
        DataFrame: DataFrame
    """
    request_info = _get_by_date_request_info(asset_class, date, repos)
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def get_distinct_in_field(asset_class: AssetClass, field: str) -> List[str]:
    """
    Get all distinct values in field.

    Args:
        asset_class (AssetClass): Asset class enum (COMMODITY, CREDIT, EQUITY, FOREX, RATES)
        field (str): Field to query for

    Returns:
        List[str]: List of all distinct values in field
    """
    request_info = _get_distinct_in_field_request_info(asset_class, field)
    return __get_request(**request_info)


@data_request
def commodity(data_filter: Optional[DataFilter] = None):
    """
    Get SDR Commodity dataset.

    Args:
        data_filter (Optional[DataFilter]): Dataset query filter

    Returns:
        DataFrame: DataFrame
    """
    request_info = _commodity_request_info(data_filter)
    return query_dataset(**request_info)


@data_request
def credit(data_filter: Optional[DataFilter] = None):
    """
    Get SDR Credit dataset.

    Args:
        data_filter (Optional[DataFilter]): Dataset query filter

    Returns:
        DataFrame: DataFrame
    """
    request_info = _credit_request_info(data_filter)
    return query_dataset(**request_info)


@data_request
def equity(data_filter: Optional[DataFilter] = None):
    """
    Get SDR Equity dataset.

    Args:
        data_filter (Optional[DataFilter]): Dataset query filter

    Returns:
        DataFrame: DataFrame
    """
    request_info = _equity_request_info(data_filter)
    return query_dataset(**request_info)


@data_request
def forex(data_filter: Optional[DataFilter] = None):
    """
    Get SDR Forex dataset.

    Args:
        data_filter (Optional[DataFilter]): Dataset query filter

    Returns:
        DataFrame: DataFrame
    """
    request_info = _forex_request_info(data_filter)
    return query_dataset(**request_info)


@data_request
def rates(data_filter: Optional[DataFilter] = None):
    """
    Get SDR Rates dataset.

    Args:
        data_filter (Optional[DataFilter]): Dataset query filter

    Returns:
        DataFrame: DataFrame
    """
    request_info = _rates_request_info(data_filter)
    return query_dataset(**request_info)



from algora.api.service.datasets.field.asynchronous import (
    async_get_field, async_create_field, async_update_field, async_delete_field
)
from algora.api.service.datasets.field.synchronous import get_field, create_field, update_field, delete_field



from algora.api.service.datasets.field_group.model import FieldGroupRequest


def _get_field_group_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/field-group/{id}"
    }


def _get_field_groups_request_info() -> dict:
    return {
        "endpoint": f"config/datasets/field-group"
    }


def _create_field_group_request_info(request: FieldGroupRequest) -> dict:
    return {
        "endpoint": f"config/datasets/field-group",
        "json": request.request_dict()
    }


def _update_field_group_request_info(id: str, request: FieldGroupRequest) -> dict:
    return {
        "endpoint": f"config/datasets/field-group/{id}",
        "json": request.request_dict()
    }


def delete_field_group_request_info(id: str) -> dict:
    return {
        "endpoint": f"config/datasets/field-group/{id}"
    }



import asyncio
from typing import Iterable, Dict, List

from algora.common.data_source.abstract import Source, AsyncSource
from algora.common.model import DataRequest
from algora.common.type import DataResponse


class DataSource:
    """
    Data source wrapper class which holds source classes and functions and performs data querying.
    """

    class _Builder:
        def __init__(self):
            self._blocking_providers = []
            self._async_providers = []

        @classmethod
        def new_instance(cls):
            return cls()

        def build(self):
            return DataSource(self._blocking_providers, self._async_providers)

        def blocking_providers(self, *providers):
            self._blocking_providers = [_provider() for _provider in providers]
            return self

        def async_providers(self, *providers):
            self._async_providers = [_provider() for _provider in providers]
            return self

    @classmethod
    def builder(cls) -> _Builder:
        return cls._Builder.new_instance()

    def __init__(
            self,
            blocking_providers: Iterable[Source],
            async_providers: Iterable[AsyncSource]
    ):
        self.blocking_providers = blocking_providers
        self.async_providers = async_providers

    async def get_data(self, request: DataRequest) -> Dict[str, DataResponse]:
        """
        Query all data sources and return dict of name to data response. Uses asynchronous sources if provided,
        else uses synchronous sources.

        Args:
            request (DataRequest): Data request

        Returns:
            Dict[str, DataResponse]: Dict of name to data response
        """
        async_provider_names = [_provider.name() for _provider in self.async_providers]
        async_data = await self.get_async_data(request)
        data = self.get_blocking_data(request, names_to_skip=async_provider_names)
        data.update(async_data)
        return data

    def get_blocking_data(self, request: DataRequest, names_to_skip: List[str] = None) -> Dict[str, DataResponse]:
        """
        Query all data sources and return dict of name to data response.

        Args:
            request (DataRequest): Data request
            names_to_skip (List[str]): List of data sources to skip when querying

        Returns:
            Dict[str, DataResponse]: Dict of name to data response
        """
        names_to_skip = [] if names_to_skip is None else names_to_skip
        return {
            _provider.name(): _provider.get_data(request)
            for _provider in self.blocking_providers if _provider.name() not in names_to_skip
        }

    async def get_async_data(self, request: DataRequest) -> Dict[str, DataResponse]:
        """
        Asynchronously query all data sources and return dict of name to data response.

        Args:
            request (DataRequest): Data request

        Returns:
            Dict[str, DataResponse]: Dict of name to data response
        """

        # TODO: Make a helper in algoralabs sdk
        async def key_value_coroutine(_provider):
            key = _provider.name()
            value = await _provider.get_data(request)
            return key, value

        coroutines = [key_value_coroutine(_provider) for _provider in self.async_providers]
        result = await asyncio.gather(*coroutines)
        return dict(result)



from typing import List, Any, Dict

from algora.api.service.documents.__util import (
    _get_document_request_info, _search_documents_request_info, _create_document_request_info,
    _update_document_request_info, _delete_document_request_info
)
from algora.api.service.documents.model import SearchDocumentRequest, DocumentRequest
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_get_request, __async_put_request, __async_post_request, __async_delete_request,
)


@async_data_request(transformer=no_transform)
async def async_get_document(id: str) -> Dict[str, Any]:
    """
    Asynchronously get documents by ID.

    Args:
        id: (str): Document ID

    Returns:
        Dict[str, Any]: Document response
    """
    request_info = _get_document_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_search_documents(request: SearchDocumentRequest) -> List[Dict[str, Any]]:
    """
    Asynchronously search all documents.

    Args:
        request: (SearchDocumentRequest): Document search request

    Returns:
        List[Dict[str, Any]]: List of documents response
    """
    request_info = _search_documents_request_info(request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_document(request: DocumentRequest) -> Dict[str, Any]:
    """
    Asynchronously create documents.

    Args:
        request: (DocumentRequest): Document request

    Returns:
        Dict[str, Any]: Document response
    """
    request_info = _create_document_request_info(request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_document(id: str, request: DocumentRequest) -> Dict[str, Any]:
    """
    Asynchronously update documents.

    Args:
        id (str): Document ID
        request: (DocumentRequest): Document request

    Returns:
        Dict[str, Any]: Document response
    """
    request_info = _update_document_request_info(id, request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_document(id: str) -> None:
    """
    Asynchronously delete documents by ID.

    Args:
        id (str): Document ID

    Returns:
        None
    """
    request_info = _delete_document_request_info(id)
    return await __async_delete_request(**request_info)
